from pprint import pprint
import sys
import os
import PySide6
from PySide6.QtWidgets import QApplication, QMainWindow, QDialog, QListWidgetItem, QMessageBox
from PySide6 import QtCore
from PySide6 import QtCharts
from mainwindow import Ui_MainWindow
from edit_dialog import Ui_Dialog
from sqlalchemy import create_engine, text
from sqlalchemy.orm import Session
from statistics import *

dirname = os.path.dirname(PySide6.__file__)
plugin_path = os.path.join(dirname, 'plugins', 'platforms')
os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = plugin_path


class EditDialog(QDialog):
    def __init__(self, regions, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)

        self.ui.btnAdd.clicked.connect(self.accept)
        self.ui.btnCancel.clicked.connect(self.reject)

        print(regions)
        for r in regions.values():
            self.ui.cmbRegions.addItem(r.title, r)

    def get_data(self):
        return {
            "region_id": self.ui.cmbRegions.currentData().id,
            "year": self.ui.txtYear.text(),
            "population": self.ui.txtPopulation.text()
        }


class UpdateDialog(EditDialog):
    def __init__(self, regions, init_data, *args, **kwargs) -> None:
        super().__init__(regions, *args, **kwargs)

        self.ui.btnAdd.setText('Изменить')
        self.ui.cmbRegions.setEnabled(False)

        region_name = regions[init_data.region_id].title
        self.ui.cmbRegions.setCurrentText(region_name)

        self.ui.txtPopulation.setText(str(init_data.population))
        self.ui.txtYear.setText(str(init_data.year))


class MainWindow(QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.engine = create_engine("sqlite+pysqlite:///mydatabase.db", echo=True)

        self.load_regions()
        self.load_years()
        self.load_population()

        self.ui.cmbRegions.currentIndexChanged.connect(self.load_population)
        self.ui.cmbYear.currentIndexChanged.connect(self.load_population)
        self.ui.btnAdd.clicked.connect(self.on_btnAdd_click)
        self.ui.btnRemove.clicked.connect(self.on_btnRemove_click)
        self.ui.btnEdit.clicked.connect(self.on_btnEdit_click)

    def on_btnEdit_click(self):
        item = self.ui.lstItems.currentItem()
        init_data = item.data(QtCore.Qt.ItemDataRole.UserRole)

        dialog = UpdateDialog(self.regions, init_data)
        r = dialog.exec()
        if r == 0:
            return
        
        data = dialog.get_data()

        with Session(self.engine)  as s:
            query = """
            UPDATE populations
            SET year = :y, population = :p
            WHERE id = :id
            """

            s.execute(text(query), {
                "y": data['year'],
                "p": data['population'],
                "id": init_data.id,
            })
            s.commit()

        self.load_population()
        self.load_years()

    def on_btnRemove_click(self):
        item = self.ui.lstItems.currentItem()
        data = item.data(QtCore.Qt.ItemDataRole.UserRole)

        r = QMessageBox.question(self, "Подтверждение", "Точно ли хотите удалить запись?")
        if r == QMessageBox.StandardButton.No:
            return

        with Session(self.engine) as s:
            query = """
            DELETE 
            FROM populations 
            WHERE id = :id
            """

            s.execute(text(query), {"id": data.id})
            s.commit()

        self.load_population()
        self.load_years()
    
    def on_btnAdd_click(self):
        dialog = EditDialog(self.regions)
        r = dialog.exec()
        if r == 0:
            return

        data = dialog.get_data()
        with Session(self.engine) as s:
            query = """
            INSERT INTO populations(region_id, year, population)
            VALUES (:rid, :y, :p)
            """

            s.execute(text(query), {
                "rid": data['region_id'],
                "y": data['year'],
                "p": data['population'],
            })
            s.commit()

        self.load_population()
        self.load_years()

    def load_population(self):
        regions_data = self.ui.cmbRegions.currentData()
        if regions_data:
            region_id = self.ui.cmbRegions.currentData().id
        else:
            region_id = 0

        year = self.ui.cmbYear.currentText()

        self.ui.lstItems.clear()
        self.rows = []

        with Session(self.engine) as s:
            query = """
            SELECT *
            FROM populations
            WHERE (:rid = 0 OR region_id = :rid) 
                AND (:y = '-' OR year = :y)
            ORDER BY  year DESC
            """

            rows = s.execute(text(query), {"rid": region_id, "y": year})

            for r in rows:
                region_name = self.regions[r.region_id].title

                item = QListWidgetItem(f"{region_name} {r.year} {r.population}")
                item.setData(QtCore.Qt.ItemDataRole.UserRole, r)
                self.ui.lstItems.addItem(item)

                self.rows.append(r)

        self.show_statistics()
        self.draw_line_chart()
        self.draw_pie_chart()
        self.draw_bar_chart()

    def draw_bar_chart(self):
        self.data_by_regions = {}

        years = set()
        for row in self.rows:
            items = self.data_by_regions.setdefault(row.region_id, {})
            items[row.year] = row.population
            years.add(row.year)

        years = sorted(years)

        series = QtCharts.QHorizontalStackedBarSeries()

        for region_id, region_data in self.data_by_regions.items():
            region_title = self.regions[region_id].title
            bar_set = QtCharts.QBarSet(region_title)
            for year in years:
                value = region_data.get(year, 0)
                bar_set.append(value)
            
            bar_set.setLabelColor("#0d0082")
            series.append(bar_set)

        series.setLabelsVisible()
        series.setLabelsPrecision(20)

        chart = QtCharts.QChart()
        chart.setAnimationOptions(QtCharts.QChart.AnimationOption.SeriesAnimations)
        chart.addSeries(series)
        chart.createDefaultAxes()

        axis = QtCharts.QBarCategoryAxis()
        axis.append([str(i) for i in years])
        chart.setAxisY(axis)
        series.attachAxis(axis)

        self.ui.chartView3.setChart(chart)


    def draw_pie_chart(self):
        self.data_by_regions = {}

        for row in self.rows:
            items = self.data_by_regions.setdefault(row.region_id, [])
            items.append(row)

        series = QtCharts.QPieSeries()
        for region_id, region_data in self.data_by_regions.items():
            region_name = self.regions[region_id].title
            avg = int(mean([i.population for i in region_data]))
            series.append(f"{region_name}", avg)
        
        series.setLabelsVisible()

        chart = QtCharts.QChart()
        chart.addSeries(series)

        self.ui.chartView2.setChart(chart)


    def draw_line_chart(self):
        self.data_by_regions = {}

        for row in self.rows:
            items = self.data_by_regions.setdefault(row.region_id, [])
            items.append(row)

        chart = QtCharts.QChart()
        for region_id, region_data in self.data_by_regions.items():
            region_name = self.regions[region_id].title

            series = QtCharts.QLineSeries()
            series.setName(region_name)
            series.setPointsVisible(True)
            series.setMarkerSize(4)
            for row in region_data:
                series.append(row.year, row.population)
            
            chart.addSeries(series)
            
        chart.createDefaultAxes()
        chart.axisX().setTitleText("Год")
        chart.axisX().setLabelFormat("%i")
        chart.axisX().setMax(chart.axisX().max() + 1)
        chart.axisX().setMin(chart.axisX().min() - 1)
        
        chart.axisY().setLabelFormat("%i")
        chart.axisY().setTitleText("Население, чел.")
        chart.axisY().setMax(chart.axisY().max() + 100000)
        chart.axisY().setMin(chart.axisY().min() - 100000)

        self.ui.chartView1.setChart(chart)

    def show_statistics(self):
        max_row = max(self.rows, key=lambda row: row.population)
        min_row = min(self.rows, key=lambda row: row.population)
        avg_population = int(mean([i.population for i in self.rows]))

        region_name = self.regions[max_row.region_id].title
        min_region_name = self.regions[min_row.region_id].title

        self.ui.lblStatistics.setText(f"""
        Наибольший показатель числености населения составил 
        <br> <b style="color: #f542c8; font-size: 16px">{max_row.population}</b> человека в  <b style="color: #4290f5">{max_row.year}</b> году в регионе <u><b>{region_name}</b></u>
        <hr>
        Наименьший показатель числености населения составил 
        <br> <b style="color: #f542c8; font-size: 16px">{min_row.population}</b> человека в <b style="color: #4290f5">{min_row.year}</b> году в регионе <u><b>{min_region_name}</b></u>
        <hr>
        Среднее значение населения составило: <span style="background: yellow">{avg_population} человека</span>
        """)



    def load_regions(self):
        self.regions = {}

        with Session(self.engine) as s:
            query = """
            SELECT *
            FROM regions
            """

            rows = s.execute(text(query))
            for r in rows:
                self.regions[r.id] = r

        self.ui.cmbRegions.addItem("-")
        for r in self.regions.values():
            self.ui.cmbRegions.addItem(r.title, r)
            
    def load_years(self):
        self.ui.cmbYear.clear()
        self.ui.cmbYear.addItem("-")

        with Session(self.engine) as s:
            query = """
            SELECT DISTINCT year
            FROM populations
            ORDER BY 1 DESC
            """

            rows = s.execute(text(query))
            for r in rows:
                self.ui.cmbYear.addItem(str(r.year))


if __name__ == "__main__":
    app = QApplication(sys.argv)

    window = MainWindow()
    window.show()

    sys.exit(app.exec())